"""
MQTTk - Lightweight graphical MQTT client and message analyser

Copyright (C) 2022  Máté Szabó

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.
"""

import tkinter as tk
import tkinter.ttk as ttk
import platform


class ScrollFrame(tk.Frame):
    """
    Borrowed from Mark Pointing
    https://gist.github.com/mp035/9f2027c3ef9172264532fcd6262f3b01
    https://github.com/mp035
    """
    def __init__(self, parent):
        super().__init__(parent)

        self.canvas = tk.Canvas(self, borderwidth=0, background="#ffffff", highlightthickness=0)
        self.viewPort = tk.Frame(self.canvas,
                                 background="#ffffff", highlightthickness=0)
        self.vsb = ttk.Scrollbar(self, orient="vertical", command=self.canvas.yview)
        self.canvas.configure(yscrollcommand=self.vsb.set)
        self.vsb.pack(side="right", fill="y")
        self.canvas.pack(side="left", fill="both", expand=True)
        self.canvas_window = self.canvas.create_window((4, 4), window=self.viewPort, anchor="nw",
                                                       tags="self.viewPort")
        self.viewPort.bind("<Configure>",
                           self.onFrameConfigure)
        self.canvas.bind("<Configure>",
                         self.onCanvasConfigure)
        self.viewPort.bind('<Enter>', self.onEnter)
        self.viewPort.bind('<Leave>', self.onLeave)
        self.onFrameConfigure(None)

    def onFrameConfigure(self, event):
        self.canvas.configure(scrollregion=self.canvas.bbox("all"))

    def onCanvasConfigure(self, event):
        canvas_width = event.width
        self.canvas.itemconfig(self.canvas_window, width=canvas_width)

    def onMouseWheel(self, event):
        if platform.system() == 'Windows':
            self.canvas.yview_scroll(int(-1 * (event.delta / 120)), "units")
        elif platform.system() == 'Darwin':
            self.canvas.yview_scroll(int(-1 * event.delta), "units")
        else:
            if event.num == 4:
                self.canvas.yview_scroll(-1, "units")
            elif event.num == 5:
                self.canvas.yview_scroll(1, "units")

    def onEnter(self, event):
        if platform.system() == 'Linux':
            self.canvas.bind_all("<Button-4>", self.onMouseWheel)
            self.canvas.bind_all("<Button-5>", self.onMouseWheel)
        else:
            self.canvas.bind_all("<MouseWheel>", self.onMouseWheel)

    def onLeave(self, event):
        if platform.system() == 'Linux':
            self.canvas.unbind_all("<Button-4>")
            self.canvas.unbind_all("<Button-5>")
        else:
            self.canvas.unbind_all("<MouseWheel>")

    def to_bottom(self):
        self.canvas.yview_moveto(1.0)



"""
MQTTk - Lightweight graphical MQTT client and message analyser

Copyright (C) 2022  Máté Szabó

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.
"""

import itertools

BEGIN_PRINTABLES = 33
END_PRINTABLES = 126


def hex_group_formatter(iterable):
    chunks = [iter(iterable)] * 4
    return '   '.join(
        ' '.join(format(x, '0>2x') for x in chunk)
        for chunk in itertools.zip_longest(*chunks, fillvalue=0))


def ascii_group_formatter(iterable):
    return ''.join(
        chr(x) if BEGIN_PRINTABLES <= x <= END_PRINTABLES else '.'
        for x in iterable)


def hex_viewer(message_data, chunk_size=16):
    header = hex_group_formatter(range(chunk_size))
    yield 'ADDRESS        {:<53}       ASCII'.format(header)
    yield ''
    template = '{:0>8x}       {:<53}       {}'
    for chunk_count in itertools.count(0):
        start = (chunk_size+1)*(chunk_count)
        finish = start + chunk_size
        if len(message_data) < finish:
            finish = len(message_data)
        chunk = message_data[start:finish]
        if not chunk:
            return
        yield template.format(
            chunk_count * chunk_size,
            hex_group_formatter(chunk),
            ascii_group_formatter(chunk))



import tkinter as tk
import tkinter.ttk as ttk
from mqttk.constants import CONNECT
from functools import partial
import os
import traceback


class BrokerStats(ttk.Frame):
    def __init__(self, master, root, log, *args, **kwargs):
        super().__init__(master=master, *args, **kwargs)

        self.root = root
        self.mqtt_manager = None
        self.log = log

        self.header_frame = ttk.Frame(self)

        self.subscribe_button = ttk.Button(self.header_frame,
                                           text="Subscribe",
                                           command=self.subscribe,
                                           state='disabled')
        self.subscribe_button.pack(side=tk.RIGHT, pady=3, padx=3)

        self.unsubscribe_button = ttk.Button(self.header_frame,
                                             text="Unsubscribe",
                                             command=self.unsubscribe,
                                             state='disabled')
        self.unsubscribe_button.pack(side=tk.RIGHT, pady=3, padx=3)

        self.header_frame.pack(fill="x", padx=3, pady=3)

        self.broker_stats_frame = ttk.Frame(self)
        self.broker_stats_frame.pack(fill='both', expand=1)
        self.broker_stats_treeview = ttk.Treeview(self.broker_stats_frame,
                                                  show="tree headings",
                                                  columns=('value',))
        self.vertical_scrollbar = ttk.Scrollbar(self.broker_stats_frame, orient="vertical",
                                                command=self.broker_stats_treeview.yview)
        self.vertical_scrollbar.pack(side=tk.RIGHT, fill="y")
        self.broker_stats_treeview.configure(yscrollcommand=self.vertical_scrollbar.set)
        self.broker_stats_treeview.pack(fill='both',
                                        expand=1)
        self.broker_stats_treeview.heading('#0', text="Statistic")
        self.broker_stats_treeview.column('#0', minwidth=300, width=300, stretch=tk.NO)
        self.broker_stats_treeview.heading("value", text="Value")
        self.broker_stats_treeview.column('value', minwidth=200, width=200, stretch=tk.NO)

    def subscribe(self, *args, **kwargs):
        try:
            callback = partial(self.on_mqtt_message, subscription_pattern='$SYS/broker/#')
            callback.__name__ = "MyCallback"  # This is to fix some weird behaviour of the paho client on linux
            self.mqtt_manager.add_subscription(topic_pattern='$SYS/broker/#',
                                               on_message_callback=callback)
        except Exception as e:
            self.log.exception("Failed to subscribe!", e)
            return

        self.subscribe_button["state"] = "disabled"
        self.unsubscribe_button["state"] = "normal"

    def unsubscribe(self, *args, **kwargs):
        try:
            self.mqtt_manager.unsubscribe('$SYS/broker/#')
        except Exception as e:
            self.log.warning("Broker stats failed to unsubscribe", "maybe a failed subscription?")
        self.unsubscribe_button["state"] = "disabled"
        self.subscribe_button["state"] = "normal"
        self.flush_messages()

    def on_mqtt_message(self, _, __, msg, subscription_pattern):
        try:
            try:
                payload_decoded = str(msg.payload.decode("utf-8"))
            except Exception:
                payload_decoded = msg.payload

            topic_split = msg.topic.split("/")[2:]

            if 1 < len(topic_split):
                if topic_split[0] not in self.broker_stats_treeview.get_children(""):
                    try:
                        self.broker_stats_treeview.insert("", "end", topic_split[0], text=topic_split[0])
                    except Exception:
                        print(msg.topic, topic_split[0], self.broker_stats_treeview.get_children(""))
                        raise
                for i in range(1, len(topic_split) - 1):
                    parent_topic = "/".join(topic_split[0:i])
                    topic = "/".join(topic_split[0:i + 1])
                    if topic not in self.broker_stats_treeview.get_children(parent_topic):
                        self.broker_stats_treeview.insert(parent_topic, "end", topic, text=topic_split[i])

            parent_topic = "/".join(topic_split[0:-1])
            topic = "/".join(topic_split)
            if topic not in self.broker_stats_treeview.get_children(parent_topic):
                self.broker_stats_treeview.insert(parent_topic,
                                                  "end",
                                                  "/".join(topic_split),
                                                  text=topic_split[-1],
                                                  values=(payload_decoded,))
            else:
                self.broker_stats_treeview.set(topic, "value", payload_decoded)

        except Exception as e:
            self.log.exception("Exception inserting new message to treeview",
                               os.linesep, msg.topic, msg.payload, os.linesep, e, traceback.format_exc())

    def interface_toggle(self, connection_state, mqtt_manager):
        # Subscribe tab items
        self.mqtt_manager = mqtt_manager
        self.subscribe_button.configure(state="normal" if connection_state is CONNECT else "disabled")

    def flush_messages(self):
        for child in self.broker_stats_treeview.get_children():
            self.broker_stats_treeview.delete(child)
